/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
  negotiate format with /dev/dsp (or /dev/dsp1 or a file) in constructor
  convert to negotiated format in put()
*/

extern "C" {
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
	   }

#include <cmath>
#include <iostream>
#include <cstddef>

using namespace std;

#include "soundout.h"
#include "dprintf.h"

soundout *
mydsp(int frag)
{
  static soundout mine(frag);

  return &mine;
}

size_t
soundout::getbufsiz()
{
  return blocksize;
}

soundout::~soundout()
{
  if (dspfd != -1)
    {
      close(dspfd);
      dspfd = -1;
    }
  if (buf1)
    delete [] buf1;
}

unsigned fragarg = 0;

soundout::soundout(int frag)
  : dspfd(-1), blocksize(0), buf1(0)
{
  dspfd = open("/dev/dsp", O_WRONLY);

  if (dspfd == -1)
    throw DeviceError();

  fragarg = frag;

  if (ioctl(dspfd, SNDCTL_DSP_SETFRAGMENT, &fragarg) == -1)
    throw DeviceError();

  dprintf(1, "/dev/dsp returned frag arg of %x\n", fragarg);

  if (ioctl(dspfd, SNDCTL_DSP_GETBLKSIZE, &blocksize) == -1)
    throw DeviceError();

  fprintf(stderr, "%d\n", blocksize);

  buf1 = new unsigned char[blocksize];

  int bits = 16;

  if (ioctl(dspfd, SNDCTL_DSP_SAMPLESIZE, &bits) == -1)
    throw DeviceError();

  int stereop = 1;

  if (ioctl(dspfd, SNDCTL_DSP_STEREO, &stereop) == -1)
    throw DeviceError();

  unsigned long ulrate = 44100;

  if (ioctl(dspfd, SNDCTL_DSP_SPEED, &ulrate) == -1)
    throw DeviceError();

  if (ioctl(dspfd, SNDCTL_DSP_SYNC, 0) != 0)
    throw DeviceError();
}

static inline short
f2s16(float f)
{
  return (short) (f * 32766);
}

void
convert2(const float *in, size_t nsamps, unsigned char *out)
{
  size_t i = 0;

  for (; i < nsamps * 2; i++)
    {
      bool negativep = false;
      float f = *in++;
      if (f < 0)
	{
	  negativep = true;
	  f = -f;
	}

      // clip f if need be
      if (f > 1)
	f = 1;

      unsigned long i = static_cast<unsigned long>(f * 32766);

      if (negativep)
	// implement definition of two's-complement
	{
	  i = (~i & 0xFFFF);
	  i += 1;
	}

      *out++ = i & 0xFF;
      *out++ = (i >> 8) & 0xFF;
    }
}

bool
soundout::put(const float *f, size_t n)
{
  convert2(f, n, buf1);
  audio_buf_info i;

#define CHECK
#ifdef CHECK
  if (ioctl(dspfd, SNDCTL_DSP_GETOSPACE, &i) != -1)
    {
      if (i.bytes == ((fragarg >> 16) * (1 << (fragarg & 0xFFFF))))
	dprintf(1, "ospace underflow!\n");
    }
#endif

  struct timeval a, b;
  gettimeofday(&a, 0);
  if (write(dspfd, buf1, n*4) != (int) (n*4))
    return false;
  gettimeofday(&b, 0);
  //printf("%d usecs\n", b.tv_usec - a.tv_usec);
  return true;
}

// kludgey..but oh well :)
static int cuefd = -1;

void
cue_put(float *f, size_t n)
{
  unsigned char *buf1 = new unsigned char[4 * n];
  convert2(f, n, buf1);
  write(cuefd, buf1, n*4);
  delete [] buf1;
}

bool
cue_try(const char *cue_channel, int frag)
{
  cuefd = open(cue_channel, O_WRONLY);
  if (cuefd == -1)
    {
      dprintf(0, "problem opening cue channel on %s: %s\n",
	      cue_channel, strerror(errno));
      return false;
    }

  int fragarg = frag;

  if (ioctl(cuefd, SNDCTL_DSP_SETFRAGMENT, &fragarg) == -1)
    {
      close(cuefd);
      cuefd = -1;
      return false;
    }

  int bits = 16;

  if (ioctl(cuefd, SNDCTL_DSP_SAMPLESIZE, &bits) == -1
      || bits != 16)
    {
      close(cuefd);
      cuefd = -1;
      return false;
    }

  int stereop = 1;

  if (ioctl(cuefd, SNDCTL_DSP_STEREO, &stereop) == -1
      || stereop != 1)
    {
      close(cuefd);
      cuefd = -1;
      return false;
    }

  unsigned long ulrate = 44100;

  if (ioctl(cuefd, SNDCTL_DSP_SPEED, &ulrate) == -1
      || ulrate != 44100)
    {
      close(cuefd);
      cuefd = -1;
      return false;
    }

  // this was hanging sometimes :(
#if 0
  if (ioctl(cuefd, SNDCTL_DSP_SYNC, 0) != 0)
    {
      close(cuefd);
      cuefd = -1;
      return false;
    }
#endif

  return true;
}

// !!! make part of soundout and modify to use buf1
void
outfile_put(FILE *f, const float *buf, size_t n)
{
}
