/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

extern "C" {
#include <pthread.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
}

#include <string>
#include <algorithm>
#include <exception>
#include <cstddef>
#include <new>

#include <iostream>

using namespace std;

struct samppair
{
  float left, right;
  
  samppair(float l, float r)
    : left(l), right(r)
    {}
  
  samppair()
    : left(0), right(0)
    {}
};

static inline samppair
interp(float factor, samppair a, samppair b)
{
  return samppair(a.left * (1 - factor) + b.left * factor,
		  a.right * (1 - factor) + b.right * factor);
}

static inline float
qinterp1(float d, float fx0, float fx1, float fx2)
{
  float fx2x1 = fx2 - fx1;
  float fx1x0 = fx1 - fx0;
  float fx2x0 = (fx2x1 - fx1x0) / 2;
    
  return fx0 + fx2x0 * d + fx2x0 * d * (d - 1);
}

static inline samppair
qinterp(float factor, samppair a, samppair b, samppair c)
{
  return samppair(qinterp1(factor, a.left, b.left, c.left),
		  qinterp1(factor, a.right, b.right, c.right));
}

class wavfilea
{
  static const size_t default_chunksize = (16 * 1024);
  static const size_t default_nchunks = 6;

  class chunk
  {
  public:

    bool validp;		// does this contain valid file data
    off_t off;			// the offset in the file
    off_t dlen; 		// length of valid data
    size_t len;			// length of chunk block
    unsigned char *p;		// chunk block

    chunk(size_t n)
      : validp(false), off(0), dlen(0), len(n)
      {
	p = new unsigned char[len];
      }

    ~chunk() { delete [] p; }
  };

  // the function executed by the reader thread
  static void *reader(void *);
  // descriptor for our reader thread
  pthread_t readert;
  // mutex to coordinate between consumer and reader
  pthread_mutex_t m;
  // condition reader waits on when there's nothing to read
  pthread_cond_t readerc;

  chunk *chunks;
  size_t chunksize;
  size_t nchunks;

  string file;
  int fd;
  
  // our phase wrt the current sample
  float phase;
  // the last, penultimate and antepenultimate sample read from the file
  samppair last, last1, last2;
  off_t p;			// location of next sample in file
  const unsigned char *pp;	// location of next sample in memory
  // sentinel for quick check of running out of data
  const unsigned char *sent;
  
  // info about the wav file
  unsigned long rate;
  int nchannels;
  // better as bits?
  int samplesize;		// in bytes
  // ULAW, ALAW, ADPCM?
  enum {SIGNED, UNSIGNED} format;
  off_t datastart;		// start of sample data in the file
  off_t datalen;		// length of the sample data

  // info about our clients needs
  size_t rate_wanted;
  size_t channels_wanted;
  bool quadraticp;

  // utility functions
  void parse_wavinfo(const unsigned char *p, int n);
  bool schedule_chunk(int &cn, off_t &fpos, off_t &rlen);
  bool
    wavfilea::run_out_of_datap()
    {
      // quickie check
      if (pp + (samplesize * nchannels) <= sent)
	return false;
      else 
	return really_run_out_of_datap();
    }
  bool really_run_out_of_datap();
  samppair next_samp();
  off_t f2off_t(float a);

  // is-there-a-loop and it's start and end points
  bool loopp;
  off_t loopa;
  off_t loopb;

  bool cueingpval, seekingpval, invalidpval;

public:

  class InvalidArgument : public exception {};
  class IllFormatted : public exception {};
  class OpenError : public exception {};

  wavfilea(string name, size_t rate = 44100, size_t nchannels = 2, 
	   bool quadraticp = false);
  ~wavfilea();
  void get(float *fbuf, size_t nsamps, float resample = 1.0);
  float tell();
  void seek(float);

  void setloop(float, float);
  void clearloop();
  bool getloop(float &, float &);

  // these should really be protected by the mutex
  bool cueingp() { return cueingpval; }
  bool seekingp() { return seekingpval; }
  bool invalidp() { return invalidpval; }
  bool goodtogetp()
    {
      return !cueingp() && !seekingp() && !invalidp();
    }

  float size();
};
